@@ -10,7 +10,7 @@ use Coro::Channel;
 use Coro::ProcessPool::Process;
 use Coro::ProcessPool::Util qw(cpu_count);
 
-our $VERSION = '0.18_2';
+our $VERSION = '0.18_3';
 
 if ($^O eq 'MSWin32') {
     die 'MSWin32 is not supported';
@@ -49,6 +49,50 @@ has pending => (
     default => sub { {} },
 );
 
+has inbox => (
+    is      => 'lazy',
+    isa     => InstanceOf['Coro::Channel'],
+);
+
+sub _build_inbox {
+  my $self = shift;
+  return Coro::Channel->new($self->max_procs);
+}
+
+has inbox_worker => (
+    is  => 'lazy',
+    isa => InstanceOf['Coro'],
+);
+
+sub _build_inbox_worker {
+    async {
+        my $self = shift;
+
+        while (1) {
+            my $task = $self->inbox->get or last;
+            my ($f, $args, $on_success, $on_error) = @$task;
+            my $msgid = $self->start_task($f, $args);
+
+            async_pool {
+                my ($self, $msgid, $on_success, $on_error) = @_;
+                my $result = eval { $self->collect_task($msgid) };
+
+                if ($@) {
+                    my $error = $@;
+                    if (ref $on_error && ref $on_error eq 'CODE') {
+                        $on_error->($error);
+                    }
+                }
+                else {
+                    if (ref $on_success && ref $on_success eq 'CODE') {
+                        $on_success->($result);
+                    }
+                }
+            } $self, $msgid, $on_success, $on_error;
+        }
+    } @_;
+}
+
 has is_running => (
     is      => 'rw',
     isa     => Bool,
@@ -56,13 +100,23 @@ has is_running => (
 );
 
 sub DEMOLISH {
-  my $self = shift;
-  $self->shutdown;
+    my $self = shift;
+    $self->shutdown;
+}
+
+sub BUILD {
+    my $self = shift;
+    $self->inbox_worker; # ensure the worker starts
 }
 
 sub shutdown {
     my $self = shift;
+
     $self->is_running(0);
+
+    $self->inbox->shutdown;
+    $self->inbox_worker->join;
+
     my $count = $self->num_procs or return;
     for (1 .. $count) {
         my $proc = $self->procs->get;
@@ -101,8 +155,10 @@ sub checkout_proc {
         return $self->start_proc;
     }
 
+    my $proc;
+
     if (!defined $timeout) {
-        return $self->procs->get;
+        $proc = $self->procs->get;
     } else {
         my $cv = AnyEvent->condvar;
 
@@ -119,17 +175,18 @@ sub checkout_proc {
             $cv->send($self->procs->get);
         } $self, $cv;
 
-        my $proc = $cv->recv;
+        $proc = $cv->recv;
 
         if ($proc) {
             $thread_timer->throw;
-            return $proc;
         } else {
             $thread_proc->cancel;
             $thread_proc->join;
             croak 'timed out waiting for available process';
         }
     }
+
+    return $proc;
 }
 
 sub start_task {
@@ -150,8 +207,9 @@ sub start_task {
     };
 
     if ($@) {
+        my $error = $@;
         $self->checkin_proc($proc);
-        croak $@;
+        croak $error;
     } else {
         return $msgid;
     }
@@ -161,7 +219,12 @@ sub collect_task {
     my ($self, $msgid) = @_;
     my $proc = $self->pending->{$msgid} || croak 'msgid not found';
     delete $self->pending->{$msgid};
-    $self->checkin_proc($proc);
+
+    async_pool {
+      my ($self, $proc) = @_;
+      $self->checkin_proc($proc)
+    } $self, $proc;
+
     return $proc->recv($msgid);
 }
 
@@ -178,39 +241,23 @@ sub map {
 }
 
 sub defer {
-    my $self  = shift;
-    my $cv    = AnyEvent->condvar;
+    my $self = shift;
+    my $cv = AnyEvent->condvar;
     my $msgid = $self->start_task(@_);
 
     async_pool {
-        my ($self, $msgid) = @_;
+        my ($self, $msgid, $cv) = @_;
         my $result = eval { $self->collect_task($msgid) };
         $cv->croak($@) if $@;
         $cv->send($result);
-    } $self, $msgid;
+    } $self, $msgid, $cv;
 
     return sub { $cv->recv };
 }
 
 sub queue {
     my ($self, $f, $args, $on_success, $on_error) = @_;
-    my $deferred = $self->defer($f, $args);
-
-    async_pool {
-        my ($deferred, $on_success, $on_error) = @_;
-        my $result = eval { $deferred->() };
-
-        if ($@) {
-            if (ref $on_error && ref $on_error eq 'CODE') {
-                $on_error->($@);
-            }
-        }
-        else {
-            if (ref $on_success && ref $on_success eq 'CODE') {
-                $on_success->($result);
-            }
-        }
-    } $deferred, $on_success, $on_error;
+    $self->inbox->put([$f, $args, $on_success, $on_error]);
 }
 
 1;
